/*******************************************************************************
 * Copyright (c) 2000, 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *******************************************************************************/

package org.eclipse.ui.internal;

import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.eclipse.jface.util.Geometry;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.ShellAdapter;
import org.eclipse.swt.events.ShellEvent;
import org.eclipse.swt.events.ShellListener;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Monitor;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.ui.IMemento;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.contexts.IContextService;
import org.eclipse.ui.internal.presentations.PresentationFactoryUtil;

public class DetachedWindow {

    private PartStack folder;

    private WorkbenchPage page;

    //Keep the state of a DetachedWindow when switching perspectives.
    private String title;

    private Rectangle bounds = new Rectangle(0,0,0,0);

    private Shell s;
    
    private boolean hideViewsOnClose = true;
    
    private ShellListener shellListener = new ShellAdapter() {
        public void shellClosed(ShellEvent e) {
            handleClose();
        }
    };
    
    private Listener resizeListener = new Listener() {
        public void handleEvent(Event event) {
            Shell shell = (Shell) event.widget;
            folder.setBounds(shell.getClientArea());
        }
    };
    
    /**
     * Create a new FloatingWindow.
     */
    public DetachedWindow(WorkbenchPage workbenchPage) {
        //super(workbenchPage.getWorkbenchWindow().getShell());
        //setShellStyle( //SWT.CLOSE | SWT.MIN | SWT.MAX | 
        //SWT.RESIZE | getDefaultOrientation());
        this.page = workbenchPage;
        folder = new ViewStack(page, false, PresentationFactoryUtil.ROLE_VIEW);
    }

    public Shell getShell() {
        return s;
    }
    
    public void create() {
        s = ((WorkbenchWindow)page.getWorkbenchWindow()).getDetachedWindowPool().allocateShell(shellListener);
        s.setData(this);
        hideViewsOnClose = true;
        if (bounds.isEmpty()) {
            Point center = Geometry.centerPoint(page.getWorkbenchWindow().getShell().getBounds());
            Point size = new Point(300, 200);
            Point upperLeft = Geometry.subtract(center, Geometry.divide(size, 2));
            
            bounds = Geometry.createRectangle(upperLeft, size); 
        }
        getShell().setBounds(bounds);

        configureShell(s);
        
        createContents(s);
        s.layout(true);
        folder.setBounds(s.getClientArea());
    }
    
    
    /**
     * Adds a visual part to this window.
     * Supports reparenting.
     */
    public void add(ViewPane part) {

        Shell shell = getShell();
        if (shell != null)
            part.reparent(shell);
        folder.add(part);
    }

    public boolean belongsToWorkbenchPage(IWorkbenchPage workbenchPage) {
        return (this.page == workbenchPage);
    }

    public boolean close() {
        hideViewsOnClose = false;
        Shell shell = getShell();
        if (shell != null) {
            shell.close();
        }
        return true;
    }
    
    /**
     * Closes this window and disposes its shell.
     */
    private boolean handleClose() {
        
        if (hideViewsOnClose) {
            List views = new ArrayList();
            collectViewPanes(views, getChildren());
            Iterator itr = views.iterator();
            while (itr.hasNext()) {
                ViewPane child = (ViewPane) itr.next();
                page.hideView(child.getViewReference());
            }
        }

        if (folder != null)
            folder.dispose();
        
//        Shell s = getShell();
        if (s != null) {
            s.removeListener(SWT.Resize, resizeListener);
            title = s.getText();
            bounds = s.getBounds();

            // Unregister this detached view as a window (for key bindings).
			final IContextService contextService = (IContextService) getWorkbenchPage()
					.getWorkbenchWindow().getWorkbench().getAdapter(
							IContextService.class);
			contextService.unregisterShell(s);

            s.setData(null);
            s = null;
        }

        return true;
    }

    /**
     * Answer a list of the view panes.
     */
    private void collectViewPanes(List result, LayoutPart[] parts) {
        for (int i = 0, length = parts.length; i < length; i++) {
            LayoutPart part = parts[i];
            if (part instanceof ViewPane) {
                result.add(part);
            }
        }
    }

    /**
     * This method will be called to initialize the given Shell's layout
     */
    protected void configureShell(Shell shell) {
        if (title != null)
            shell.setText(title);
        shell.addListener(SWT.Resize, resizeListener);

        // Register this detached view as a window (for key bindings).
		final IContextService contextService = (IContextService) getWorkbenchPage()
				.getWorkbenchWindow().getWorkbench().getAdapter(
						IContextService.class);
        contextService.registerShell(shell,
                IContextService.TYPE_WINDOW);

        page.getWorkbenchWindow().getWorkbench().getHelpSystem().setHelp(shell,
				IWorkbenchHelpContextIds.DETACHED_WINDOW);
    }

    /**
     * Override this method to create the widget tree that is used as the window's contents.
     */
    protected Control createContents(Composite parent) {
        // Create the tab folder.
        folder.createControl(parent);

        // Reparent each view in the tab folder.
        Vector detachedChildren = new Vector();
        collectViewPanes(detachedChildren, getChildren());
        Enumeration itr = detachedChildren.elements();
        while (itr.hasMoreElements()) {
            LayoutPart part = (LayoutPart) itr.nextElement();
            part.reparent(parent);
        }

        // Return tab folder control.
        return folder.getControl();
    }

    public LayoutPart[] getChildren() {
        return folder.getChildren();
    }

    public WorkbenchPage getWorkbenchPage() {
        return this.page;
    }

    /**
     * @see IPersistablePart
     */
    public void restoreState(IMemento memento) {
        // Read the title.
        title = memento.getString(IWorkbenchConstants.TAG_TITLE);

        // Read the bounds.
        Integer bigInt;
        bigInt = memento.getInteger(IWorkbenchConstants.TAG_X);
        int x = bigInt.intValue();
        bigInt = memento.getInteger(IWorkbenchConstants.TAG_Y);
        int y = bigInt.intValue();
        bigInt = memento.getInteger(IWorkbenchConstants.TAG_WIDTH);
        int width = bigInt.intValue();
        bigInt = memento.getInteger(IWorkbenchConstants.TAG_HEIGHT);
        int height = bigInt.intValue();
        bigInt = memento.getInteger(IWorkbenchConstants.TAG_FLOAT);

        // Set the bounds.
        bounds = new Rectangle(x, y, width, height);
        if (getShell() != null) {
            getShell().setText(title);
            getShell().setBounds(bounds);
        }
        
        // Create the folder.
        IMemento childMem = memento.getChild(IWorkbenchConstants.TAG_FOLDER);
        if (childMem != null)
            folder.restoreState(childMem);
    }

    /**
     * @see IPersistablePart
     */
    public void saveState(IMemento memento) {
        if (getShell() != null) {
            title = getShell().getText();
            bounds = getShell().getBounds();
        }
        // Save the title.
        memento.putString(IWorkbenchConstants.TAG_TITLE, title);

        // Save the bounds.
        memento.putInteger(IWorkbenchConstants.TAG_X, bounds.x);
        memento.putInteger(IWorkbenchConstants.TAG_Y, bounds.y);
        memento.putInteger(IWorkbenchConstants.TAG_WIDTH, bounds.width);
        memento.putInteger(IWorkbenchConstants.TAG_HEIGHT, bounds.height);

        // Save the views.	
        IMemento childMem = memento.createChild(IWorkbenchConstants.TAG_FOLDER);
        folder.saveState(childMem);
    }
    
    /* (non-Javadoc)
     * @see org.eclipse.ui.internal.IWorkbenchDragDropPart#getControl()
     */
    public Control getControl() {
        return folder.getControl();
    }
    
    /**
     * 
     * Returns true iff the given rectangle is located in the client area of any
     * monitor.
     * 
     * @param someRectangle a rectangle in display coordinates (not null)
     * @return true iff the given point can be seen on any monitor
     */
    private static boolean intersectsAnyMonitor(Display display,
            Rectangle someRectangle) {
        Monitor[] monitors = display.getMonitors();

        for (int idx = 0; idx < monitors.length; idx++) {
            Monitor mon = monitors[idx];

            if (mon.getClientArea().intersects(someRectangle)) {
                return true;
            }
        }

        return false;
    }

    /**
     * Opens the detached window.
     */
	public int open() { 
		 
		if (getShell() == null) 
			create(); 
		 
		Rectangle bounds = getShell().getBounds(); 
        getShell().setVisible(true);
		 
		if (!bounds.equals(getShell().getBounds())) { 
			getShell().setBounds(bounds); 
		} 
		 
		return Window.OK; 
	} 
    
}
